"""Tjost configuration module"""

import click
import git
import jsonschema
import jsonschema_default
import logging
import os
import pathlib
import yaml
from tjost.helpers import deep_merge

logger = logging.getLogger(__name__)

SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {
        "defaults": {
            "type": "object",
            "properties": {
                "provider": {
                    "type": "string",
                    "enum": [
                        "libvirt",
                        "vmware",
                    ],
                    "default": "libvirt",
                },
                "memory": {
                    "type": "integer",
                    "default": 2048,
                },
                "cpus": {
                    "type": "integer",
                    "default": 2,
                },
            },
        },
        "logging": {
            "type": "object",
            "properties": {
                "journal": {
                    "type": "object",
                    "properties": {
                        "enabled": {
                            "type": "boolean",
                            "default": False,
                        },
                        "level": {
                            "type": "integer",
                            "minimum": 1,
                            "maximum": 3,
                            "default": 3,
                        },
                    },
                },
            },
        },
    },
}


class Config:
    """Tjost configuration class"""

    data: dict
    path: click.Path
    file_required: bool = False

    def __init__(self, path: click.Path = None):
        """
        Initialize the Config class.

        Automatically searches for a configuration file starting in the working directory.
        If the directory is part of a git repository, a file named `.tjost.yaml` in the
        repository root is used. If the directory is not part of a git repository, the same
        file is looked for in the current working directory.

        If a path is provided and the file does not exist, the method exits with an error.
        Otherwise only the content of the provided file is used.

        :param path: Path to the configuration file.
        """
        self.discover_config_path(path)
        self.load_config()
        logger.debug(f"Configuration loaded: {self.data}")

    def get(self, key: str = None) -> any:
        """
        Get a configuration value by key.

        :param key: The configuration key to retrieve the value from. If None, returns the entire configuration.
        :return: The configuration value.
        """

        def traverse(data: dict, keys: list) -> any:
            """
            Traverse nested dictionary to get the value for a given key path.

            :param data: The dictionary to traverse.
            :param keys: List of keys representing the path to traverse.
            :return: The value at the specified key path or the default value.
            """
            if len(keys) > 1:
                data = data.get(keys[0], {})
                return traverse(data, keys[1:])
            return data.get(keys[0])

        if key is None:
            return self.data
        return traverse(self.data, key.split("."))

    def set(self, key: str, value: any):
        """
        Set a configuration value by key.

        :param key: The configuration key to set the value for.
        :param value: The value to set.
        """
        self.data[key] = value
        try:
            jsonschema.validate(instance=self.data, schema=SCHEMA)
        except jsonschema.ValidationError as e:
            logger.error(f"Configuration data is invalid: {e}")
            raise SystemExit(1)
        self.save_config()

    def discover_config_path(self, path: click.Path = None):
        """
        Discover the configuration file path if the provided path is None. Otherwise return the provided path.

        :param path: Optional path to the configuration file.
        :return: Path to the configuration file.
        """
        if path is not None:
            self.file_required = True
            self.path = pathlib.Path(path)
            return
        try:
            # Test environments usually have a git repository
            logger.debug("Discovering git repository root")
            repo = git.Repo(search_parent_directories=True)
            discover_path = pathlib.Path(f"{repo.working_tree_dir}/.tjost.yaml")
            logger.debug(f"Found git repository root at {repo.working_tree_dir}")
        except git.exc.InvalidGitRepositoryError:
            # If we don't find a git repo, use the current working directory
            logger.debug("Falling back to current working directory")
            discover_path = pathlib.Path(f"{os.getcwd()}/.tjost.yaml")
        self.path = discover_path

    def load_config(self):
        """
        Load configuration from a file.
        """
        logger.info(f"Loading configuration from {self.path}")
        try:
            with click.open_file(self.path, "r") as f:
                data = yaml.full_load(f)
                jsonschema.validate(instance=data, schema=SCHEMA)
        except jsonschema.ValidationError as e:
            logger.error(f"Configuration file {self.path} is invalid: {e}")
            raise SystemExit(1)
        except FileNotFoundError:
            msg = f"Configuration file {self.path} not found"
            if self.file_required:
                logger.error(msg)
                raise SystemExit(1)
            logger.warning(msg)
            data = {}
        default_data = jsonschema_default.create_from(SCHEMA, {})
        self.data = deep_merge(default_data, data)

    def save_config(self):
        """
        Save configuration to a file.
        """
        logger.info(f"Saving configuration to {self.path}")
        with click.open_file(self.path, "w") as f:
            yaml.dump(self.data, f)



"""Tjost - Virtual Machine Setup Manager"""

import click
import json
import logging
from importlib.metadata import version
from tjost.config import Config
from tjost.logging_manager import LoggingManager


logger = logging.getLogger(__name__)


@click.group(context_settings={"auto_envvar_prefix": "TJOST"})
@click.pass_context
@click.option(
    "-c",
    "--config",
    "config_path",
    help="Path to configuration file.",
    type=click.Path(),
)
@click.option(
    "-n",
    "--non-interactive",
    default=False,
    help="Run in non-interactive mode and skip user inputs.",
    is_flag=True,
)
@click.option(
    "-v",
    "--verbose",
    "verbosity",
    count=True,
    help="Enable verbose output. Multiple -v options (max 3) increase the verbosity.",
)
def cli(
    ctx: click.Context,
    config_path: click.Path,
    non_interactive: bool,
    verbosity: int,
):
    """Tjost - Easily manage virtual machine setups for test environments."""
    logging_manager = LoggingManager(verbosity=verbosity)
    config = Config(path=config_path)

    journal_enabled: bool = config.get("logging.journal.enabled")
    if journal_enabled:
        journal_log_level: int = config.get("logging.journal.level")
        logging_manager.add_journal_handler(verbosity=journal_log_level)

    ctx.ensure_object(dict)
    ctx.obj["config"] = config


@cli.command("doctor")
@click.pass_context
def cli_doctor(ctx: click.Context):
    """Check the Tjost installation for potential issues."""
    click.echo("Running Tjost doctor... (not implemented)")


@cli.command("version")
@click.pass_context
@click.option(
    "-r",
    "--raw",
    "is_raw",
    is_flag=True,
    help="Show raw version string.",
    default=False,
)
def cli_version(ctx: click.Context, is_raw: bool):
    """Show the version of the Tjost CLI."""
    if ctx.parent.params.get("non_interactive"):
        is_raw = True
    if is_raw:
        click.echo(version("tjost"))
    else:
        click.echo(f"Tjost version {version('tjost')}")


# -----------------------------------------------------------------------------------------------------------------------
# Configuration Subcommands
# -----------------------------------------------------------------------------------------------------------------------
@cli.group("config")
def cli_config():
    """Manage the Tjost configuration."""
    pass


@cli_config.command("show")
@click.pass_context
def config_show(ctx: click.Context):
    """Show current configuration."""
    config = ctx.obj["config"]
    click.echo(json.dumps(config.get(), indent=2))


@cli_config.command("set")
@click.pass_context
@click.argument("key")
@click.argument("value")
def config_set(ctx: click.Context, key: str, value: str):
    """Set a configuration value."""
    config = ctx.obj["config"]
    config.set(key, value)


@cli_config.command("get")
@click.pass_context
@click.argument("key")
def config_get(ctx: click.Context, key: str):
    """Get a configuration value."""
    config = ctx.obj["config"]
    click.echo(config.get(key))



"""Logging manager for configuring and handling log outputs"""

import logging
from importlib.metadata import metadata


logger = logging.getLogger(__name__)

logging.addLevelName(logging.CRITICAL, "Crit")
logging.addLevelName(logging.DEBUG, "Debug")
logging.addLevelName(logging.ERROR, "Error")
logging.addLevelName(logging.INFO, "Info")
logging.addLevelName(logging.WARNING, "Warn")


class LoggingManager:
    """Manages logging configuration"""

    __format: str = "[{levelname:<5}] {message} ({name}:{lineno})"
    __formatter: logging.Formatter = logging.Formatter(__format, style="{")
    __verbosity: dict = {
        0: logging.ERROR,
        1: logging.WARNING,
        2: logging.INFO,
    }

    def __init__(self, verbosity: int):
        """
        Initialize logging with the specified verbosity level.
        Automatically adds a stream handler for console output.

        The verbosity levels are mapped as follows:
        - 0: ERROR
        - 1: WARNING
        - 2: INFO
        - 3 or higher: DEBUG

        :param verbosity: Verbosity level
        """
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(self.__verbosity.get(verbosity, logging.DEBUG))
        stream_handler.setFormatter(self.__formatter)
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)
        root_logger.addHandler(stream_handler)
        logger.info("Logging initialized")
        logger.info(f"Stream handler with verbosity {verbosity} added to logging")

    def add_journal_handler(self, verbosity: int):
        """
        Add a systemd journal handler with the specified verbosity level.
        The verbosity parameter behaves the same way as in the constructor.

        :param verbosity: Verbosity level
        """
        try:
            from systemd.journal import JournalHandler
        except ImportError as e:
            logger.warning("Failed to import JournalHandler from systemd.journal")
            logger.warning("Systemd journal logging will not be available")
            logger.debug(f"ImportError details: {e}")
            return

        journal_handler = JournalHandler(SYSLOG_IDENTIFIER=metadata("tjost")["Name"])
        journal_handler.setLevel(self.__verbosity.get(verbosity, logging.DEBUG))
        journal_handler.setFormatter(self.__formatter)
        root_logger = logging.getLogger()
        root_logger.addHandler(journal_handler)
        logger.info(
            f"Journal handler with verbosity level {verbosity} added to logging"
        )



"""Tjost helper functions"""


def deep_merge(dict1, dict2) -> dict:
    """
    Merge two dictionaries, including nested dictionaries.

    :param dict1: The first dictionary.
    :param dict2: The second dictionary.
    :return: The merged dictionary.
    """
    result = dict1.copy()
    for key, value in dict2.items():
        if key in result:
            if isinstance(result[key], dict) and isinstance(value, dict):
                result[key] = deep_merge(result[key], value)
            else:
                result[key] = value
        else:
            result[key] = value
    return result



"""Tjost smoke tests"""

from click.testing import CliRunner
from tjost.main import cli


def test_cli_version():
    """Test the 'tjost version' command."""
    runner = CliRunner()
    result = runner.invoke(cli, ["version"])
    assert result.exit_code == 0
    assert "Tjost version" in result.output


def test_cli_version_raw():
    """Test the 'tjost version --raw' command."""
    runner = CliRunner()
    result = runner.invoke(cli, ["version", "--raw"])
    assert result.exit_code == 0
    assert "Tjost version" not in result.output  # Raw output should not have prefix


def test_cli_doctor():
    """Test the 'tjost doctor' command."""
    runner = CliRunner()
    result = runner.invoke(cli, ["doctor"])
    assert result.exit_code == 0
    assert "Running Tjost doctor" in result.output


def test_cli_config_show():
    """Test the 'tjost config show' command."""
    runner = CliRunner()
    result = runner.invoke(cli, ["config", "show"])
    assert result.exit_code == 0
    assert "{" in result.output  # Basic check for JSON output


def test_cli_config_set_and_get():
    """Test the 'tjost config set' and 'tjost config get' commands."""
    runner = CliRunner()
    set_result = runner.invoke(cli, ["config", "set", "test_key", "test_value"])
    assert set_result.exit_code == 0

    get_result = runner.invoke(cli, ["config", "get", "test_key"])
    assert get_result.exit_code == 0
    assert "test_value" in get_result.output


def test_cli_explicit_config_file_missing():
    """Test behavior when an explicit config file is missing."""
    runner = CliRunner()
    result = runner.invoke(
        cli, ["--config", "/non_existent_config.yaml", "config", "show"]
    )
    assert result.exit_code != 0
    assert "not found" in result.output
